#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include "CImg.h"
#include "read_parameter.h"

using namespace cimg_library;
using namespace std;

void separate();
void update();

int max_iter;
double alpha;
double beta;
double Gamma;
int mx;
int my;
vector<double> cell;
vector<double> R_cell; // cellules réceptrices
vector<double> NR_cell; // cellules non-réceptrices
int dx[] = {-1, 0, -1, 1, 0, 1};
int dy[] = {-1, -1, 0, 0, 1, 1};// 6 directions de croissance

int main()
{
  ifstream fin("para.in");
  fin >> "alpha" >> alpha;
  fin >> "beta" >> beta;
  fin >> "gamma" >> Gamma;
  fin >> "size" >> mx;
  fin >> "iterations" >> max_iter;
  fin.close();
  
  my = mx;
  
  // on attribue la valeur de fond 'beta' à toutes les cellules
  cell.assign(mx*my, beta);
  R_cell.assign(mx*my, beta);
  NR_cell.assign(mx*my, beta);
  
  cell[my/2 +my*(mx/2)] = 1.0; // initialisation du cristal au centre
  
  // boucle de croissance
  for(int i=0; i<max_iter; i++){
    if(i%10 == 0) std::cout << "step " << i << " of " << max_iter << endl;
    separate();
    update();
  }
  
  // représentation graphique de l'état final du cristal
  
  double cmax = 0.0;
  for(auto& x : cell) if(x > cmax) cmax = x;
  
  double an45 = -M_PI/4.0;
  double sn45 = sin(an45);
  double cs45 = cos(an45);
  double scale = sqrt(3.0);
  int ox = mx;
  int oy = my;
  
  int fac = 2;
  int imgy = fac*my;
  int imgx = fac*mx;
  CImg<unsigned char> img(imgx,imgy,1,1);
  
  double norm = 255/cmax;
  for(int ky = 0; ky<imgy; ky++){
    for(int kx = 0; kx<imgx; kx++){
      // rotation et dilatation vers une base hexagonale plutôt que carrée
      double tx = kx - ox;
      double ty = (ky - oy) * scale;
      double tx0 = tx * cs45 - ty * sn45 + ox;
      ty = tx * sn45 + ty * cs45 + oy;
      tx = tx0;
      if(tx > 0 && tx <= imgx && ty > 0 && ty <= imgy){
        double c = cell[(int)floor(my*ty/imgy)+my*(int)floor(mx*tx/imgx)];
        if(c < 1.0 ) img(kx, ky) = 0;
        else img(kx, ky) = 255 - (unsigned char)((0.5*c*norm));
//        else img(kx, ky) = (unsigned char)((c*norm));
      }
    }
  }
  ostringstream sout;
  sout << "flocon_" << alpha << "_" << beta << "_" << Gamma << ".png";
  img.save_imagemagick_external(sout.str().c_str());
}

/**
Cette fonction sépare les cellules réceptrices des cellules non réceptrices
*/
void separate()
{
  for(int iy=0; iy<my; iy++){
    for(int ix=0; ix<mx; ix++){
      bool receptive = false;
      if(cell[iy+my*ix] >= 1.0) receptive = true;
      else{ // on vérifie les 6 voisins
        for(int j=0; j<6; j++){
          int jx = (ix + dx[j]+mx)%mx;
          int jy = (iy + dy[j]+my)%my;
          if (cell[jy%my+my*(jx%mx)] >= 1.0){
            receptive = true;
            break;
          }
        }
      }
      if (receptive){
        // on copy les cellules réceptrices dans R_cell
        // et alors NR_cell vaut zéro
        R_cell[iy+my*ix] = cell[iy+my*ix] + Gamma;
        NR_cell[iy+my*ix] = 0.0;
      }
      else{
        // on copy les cellules non réceptrices dans NR_cell
        // et alors R_cell vaut zéro
        R_cell[iy+my*ix] = 0.0;
        NR_cell[iy+my*ix] = cell[iy+my*ix];
      }
    }
  }
}


/**
Cette function diffuse les molécule d'eau parmi les cellules non réceptrices
'alpha' est le coefficient de diffusion. Le tableau 'cell' est mis à jour.
*/
void update()
{
  for(int iy=0; iy<my; iy++){
    for(int ix=0; ix<mx; ix++){
      double wsum = NR_cell[iy+my*ix] * (1.0 - alpha * 6.0 / 12.0);
      for(int j=0; j<6; j++){
        int jx = (ix + dx[j]+mx)%mx;
        int jy = (iy + dy[j]+my)%my;
        wsum += NR_cell[jy%my+my*(jx%mx)] * alpha / 12.0;
      }
      cell[iy+my*ix] = R_cell[iy+my*ix] + wsum;
    }
  }
}




