using Images, FileIO

imgx = 500
imgy = 500 # image size (500)
maxIt = 1000 # of growth steps (1000)
# snowflake will differ depending on values of these parameters:
alpha = 2
beta = 0.4
gamma = 0.001
mx = 250
my = 250 # width and height of 2DCA
ca = beta*ones(mx, my)
caRep = beta*ones(mx, my) # receptive cells
caNRep = beta*ones(mx, my) # non-receptive cells
dx = [-1 0 -1 1 0 1]
dy = [-1 -1 0 0 1 1] # 6 directions to grow
# these are for coloring the image

function separate()
    for iy = 1:my
        for ix = 1:mx
            receptive = false
            if ca[iy, ix] >= 1.0 # ice
                receptive = true
            else # check neighbors
                for j = 1:6
                    jx = ix + dx[j]
                    jy = iy + dy[j]
                    if jx > 0 && jx < mx && jy > 0 && jy < my
                        if ca[jy, jx] >= 1.0 # ice
                            receptive = true
                            break
                        end
                    end
                end
            end
            if receptive
                caRep[iy, ix] = ca[iy, ix] + gamma
                caNRep[iy, ix] = 0.0
            else
                caRep[iy, ix] = 0.0
                caNRep[iy, ix] = ca[iy, ix]
            end
        end
    end
end

function update()
    for iy = 1:my
        for ix = 1:mx
            wsum = caNRep[iy, ix] * (1.0 - alpha * 6.0 / 12.0)
            for j = 1:6 # neighbors
                jx = ix + dx[j]
                jy = iy + dy[j]
                if jx > 0 && jx < mx && jy > 0 && jy < my
                    wsum += caNRep[jy, jx] * alpha / 12.0
                end
            end
            ca[iy, ix] = caRep[iy, ix] + wsum
        end
    end
end

ca[div(my,2), div(mx,2)] = 1.0 # ice seed

for i = 1:maxIt # growth steps
    i%10 == 0 ? @printf("step: %d of %d\n" , i, maxIt) :
    # separate the array into receptive and non-receptive arrays
    separate()
    # new array: weighed averages of the non-receptive array + receptive array
    update()
end
# paint final state of the snowflake

an45 = -pi/4.0
sn45 = sin(an45)
cs45 = cos(an45)
scale = sqrt(3.0)
ox = imgx/2
oy = imgy/2

pixels = zeros(imgx, imgy)
for ky = 1:imgy
    for kx = 1:imgx
        # apply geometric transformation (scaling and rotation)
        tx = kx - ox
        ty = (ky - oy) * scale
        tx0 = tx * cs45 - ty * sn45 + ox
        ty = tx * sn45 + ty * cs45 + oy
        tx = tx0
        if tx > 0 && tx <= imgx && ty > 0 && ty <= imgy
            c = ca[floor(UInt8, my*ty/imgy+1), floor(UInt8, mx*tx/imgx+1)]
            if c >= 1.0 # ice
                pixels[kx, ky] = (c - 1.0) * 255
            end
        end
    end
end
img = colorview(Gray, pixels)
img = map(clamp01nan, img)
save("flocon.png", img)
